use nu_plugin::{EngineInterface, EvaluatedCall, SimplePluginCommand};
use nu_protocol::{
    record, Category, Example, IntoSpanned, LabeledError, PipelineData, Signature, Spanned, Value,
};

use crate::OncePlugin;

pub struct Scenario;

impl SimplePluginCommand for Scenario {
    type Plugin = OncePlugin;

    fn name(&self) -> &str {
        "once scenario"
    }

    fn usage(&self) -> &str {
        "Manage once scenarios on local or remote servers."
    }

    fn extra_usage(&self) -> &str {
        r"This command shows all available scenarios. To manage these scenarios use the sub commands, e.g.

  > once scenario.deploy"
    }

    fn signature(&self) -> Signature {
        Signature::build(self.name()).category(Category::Experimental)
    }

    fn search_terms(&self) -> Vec<&str> {
        vec!["once", "scenario"]
    }

    fn examples(&self) -> Vec<Example> {
        vec![Example {
            example: "once scenario",
            description: "List all scenarios",
            result: Some(Value::test_list(vec![Value::test_record(record! {
                "name" => Value::test_string("test"),
                "filename" => Value::test_string("test.scenario.env")
            })])),
        }]
    }

    fn run(
        &self,
        _plugin: &OncePlugin,
        engine: &EngineInterface,
        call: &EvaluatedCall,
        _input: &Value,
    ) -> Result<Value, LabeledError> {
        let head = call.head;

        const SCENARIO_SOURCE_PATH: &str = "/workspaces/Once.2023/Scenarios";
        let name: Spanned<String> = Spanned {
            span: head,
            item: "glob".to_owned(),
        };
        let decl_id = engine.find_decl(&name.item)?.ok_or_else(|| {
            LabeledError::new(format!("Can't find `{}`", name.item))
                .with_label("not in scope", name.span)
        })?;

        let mut new_call = EvaluatedCall::new(call.head);
        //new_call.add_flag("l".into_spanned(head));
        new_call.add_positional(Value::string(
            format!("{}/**/*.{{scenario,env}}", SCENARIO_SOURCE_PATH),
            head,
        ));
        // TODO: Filter out localhost is not working yet
        new_call.add_named(
            "exclude".into_spanned(head),
            Value::string("[**/localhost/*]", head),
        );

        let result: PipelineData =
            engine.call_decl(decl_id, new_call, PipelineData::Empty, true, false)?;

        let mut list = vec![];
        for res in result.into_iter() {
            list.push(res);
        }

        Ok(Value::list(list, head))
    }
}

#[test]
fn test_examples() -> Result<(), nu_protocol::ShellError> {
    use nu_plugin_test_support::PluginTest;
    PluginTest::new("once", OncePlugin.into())?.test_command_examples(&Scenario)
}
